/*
 * Copyright 2012 htfv (Aliaksei Lahachou)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.github.htfv.utils.el.internal;

import java.lang.reflect.Method;

import javax.el.FunctionMapper;

import org.apache.commons.lang3.StringUtils;

/**
 * Implements Build Configurator specific function mapper. It allows using the
 * following functions in EL:
 * <ul>
 * <li>{@code str:concat(str1, str2)} - concatenates 2 strings.</li>
 * <li>{@code str:concat3(str1, str2, str3)} - concatenates 3 strings.</li>
 * <li>{@code str:concat4(str1, str2, str3, str4)} - concatenates 4 strings.</li>
 * <li>{@code str:concat5(str1, str2, str3, str4, str5)} - concatenates 5
 * strings.</li>
 * <li>{@code str:join(str1, str2, separator)} - concatenates 2 strings with a
 * separator between them.</li>
 * <li>{@code str:join3(str1, str2, str3, separator)} - concatenates 3 strings
 * with a separator between them.</li>
 * <li>{@code str:join4(str1, str2, str3, str4, separator)} - concatenates 4
 * strings with a separator between them.</li>
 * <li>{@code str:join5(str1, str2, str3, str4, str5, separator)} - concatenates
 * 5 strings with a separator between them.</li>
 * </ul>
 *
 * @author htfv (Aliaksei Lahachou)
 */
public class DefaultFunctionMapper extends FunctionMapper
{
    /**
     * Caches reference to the {@link DefaultFunctionMapper}
     * singleton.
     *
     * @author htfv (Aliaksei Lahachou)
     */
    private static class Singleton
    {
        /**
         * Reference to the {@link DefaultFunctionMapper} singleton.
         */
        private static final FunctionMapper INSTANCE = new DefaultFunctionMapper();
    }

    /**
     * Caches reference to the {@link DefaultFunctionMapper#strConcat}
     * method.
     *
     * @author htfv (Aliaksei Lahachou)
     */
    private static class StrConcat
    {
        /**
         * Reference to the {@link DefaultFunctionMapper#strConcat}
         * method.
         */
        private static final Method INSTANCE = getMethod("strConcat",
                String.class, String.class);
    }

    /**
     * Caches reference to the
     * {@link DefaultFunctionMapper#strConcat3} method.
     *
     * @author htfv (Aliaksei Lahachou)
     */
    private static class StrConcat3
    {
        /**
         * Reference to the {@link DefaultFunctionMapper#strConcat3}
         * method.
         */
        private static final Method INSTANCE = getMethod("strConcat3",
                String.class, String.class, String.class);
    }

    /**
     * Caches reference to the
     * {@link DefaultFunctionMapper#strConcat4} method.
     *
     * @author htfv (Aliaksei Lahachou)
     */
    private static class StrConcat4
    {
        /**
         * Reference to the {@link DefaultFunctionMapper#strConcat4}
         * method.
         */
        private static final Method INSTANCE = getMethod("strConcat4",
                String.class, String.class, String.class, String.class);
    }

    /**
     * Caches reference to the
     * {@link DefaultFunctionMapper#strConcat5} method.
     *
     * @author htfv (Aliaksei Lahachou)
     */
    private static class StrConcat5
    {
        /**
         * Reference to the {@link DefaultFunctionMapper#strConcat5}
         * method.
         */
        private static final Method INSTANCE = getMethod("strConcat5",
                String.class, String.class, String.class, String.class, String.class);
    }

    /**
     * Caches reference to the {@link DefaultFunctionMapper#strJoin}
     * method.
     *
     * @author htfv (Aliaksei Lahachou)
     */
    private static class StrJoin
    {
        /**
         * Reference to the {@link DefaultFunctionMapper#strJoin}
         * method.
         */
        private static final Method INSTANCE = getMethod("strJoin",
                String.class, String.class, String.class);
    }

    /**
     * Caches reference to the {@link DefaultFunctionMapper#strJoin3}
     * method.
     *
     * @author htfv (Aliaksei Lahachou)
     */
    private static class StrJoin3
    {
        /**
         * Reference to the {@link DefaultFunctionMapper#strJoin3}
         * method.
         */
        private static final Method INSTANCE = getMethod("strJoin3",
                String.class, String.class, String.class, String.class);
    }

    /**
     * Caches reference to the {@link DefaultFunctionMapper#strJoin4}
     * method.
     *
     * @author htfv (Aliaksei Lahachou)
     */
    private static class StrJoin4
    {
        /**
         * Reference to the {@link DefaultFunctionMapper#strJoin4}
         * method.
         */
        private static final Method INSTANCE = getMethod("strJoin4",
                String.class, String.class, String.class, String.class, String.class);
    }

    /**
     * Caches reference to the {@link DefaultFunctionMapper#strJoin5}
     * method.
     *
     * @author htfv (Aliaksei Lahachou)
     */
    private static class StrJoin5
    {
        /**
         * Reference to the {@link DefaultFunctionMapper#strJoin5}
         * method.
         */
        private static final Method INSTANCE = getMethod("strJoin5",
                String.class, String.class, String.class, String.class, String.class, String.class);
    }

    /**
     * Returns reference to the {@link DefaultFunctionMapper}
     * singleton. Because {@code BuildConfiguratorFunctionMapper} does not have
     * state, it may be reused without restrictions.
     *
     * @return reference to the {@link DefaultFunctionMapper}
     *         singleton.
     */
    public static FunctionMapper getInstance()
    {
        return Singleton.INSTANCE;
    }

    /**
     * Returns a {@link Method} object which reflects a public method of the
     * {@link DefaultFunctionMapper} class.
     *
     * @param name
     *            the name of the method.
     * @param parameterTypes
     *            the list of parameters.
     *
     * @return the {@code Method} object that matches the specified {@code name}
     *         and {@code parameterTypes} or {@code null} if method was not
     *         found.
     */
    private static Method getMethod(final String name, final Class<?>... parameterTypes)
    {
        try
        {
            return DefaultFunctionMapper.class.getMethod(name, parameterTypes);
        }
        catch (NoSuchMethodException e)
        {
            return null;
        }
    }

    /**
     * Joins array of strings with a separator.
     *
     * @param separator
     *            the separator string to use.
     * @param strings
     *            the array of strings to join.
     *
     * @return the joined string.
     */
    private static String join(final String separator, final String... strings)
    {
        StringBuilder joinedString      = new StringBuilder();
        String        theRightSeparator = StringUtils.EMPTY;

        for (String str : strings)
        {
            if (StringUtils.isNotEmpty(str))
            {
                joinedString
                        .append(theRightSeparator)
                        .append(str);

                theRightSeparator = separator;
            }
        }

        return joinedString.toString();
    }

    /**
     * Implements {@code str:concat(str1, str2)} EL function - concatenates 2
     * strings.
     *
     * @param str1
     *            the first string.
     * @param str2
     *            the second string.
     *
     * @return the concatenated string.
     */
    public static String strConcat(final String str1, final String str2)
    {
        return StringUtils.join(str1, str2);
    }

    /**
     * Implements {@code str:concat3(str1, str2, str3)} EL function -
     * concatenates 3 strings.
     *
     * @param str1
     *            the first string.
     * @param str2
     *            the second string.
     * @param str3
     *            the third string.
     *
     * @return the concatenated string.
     */
    public static String strConcat3(final String str1, final String str2, final String str3)
    {
        return StringUtils.join(str1, str2, str3);
    }

    /**
     * Implements {@code str:concat4(str1, str2, str3, str4)} EL function -
     * concatenates 4 strings.
     *
     * @param str1
     *            the first string.
     * @param str2
     *            the second string.
     * @param str3
     *            the third string.
     * @param str4
     *            the fourth string.
     *
     * @return the concatenated string.
     */
    public static String strConcat4(final String str1, final String str2, final String str3,
            final String str4)
    {
        return StringUtils.join(str1, str2, str3, str4);
    }

    /**
     * Implements {@code str:concat4(str1, str2, str3, str4, str5)} EL function
     * - concatenates 5 strings.
     *
     * @param str1
     *            the first string.
     * @param str2
     *            the second string.
     * @param str3
     *            the third string.
     * @param str4
     *            the fourth string.
     * @param str5
     *            the fifth string.
     *
     * @return the concatenated string.
     */
    public static String strConcat5(final String str1, final String str2, final String str3,
            final String str4, final String str5)
    {
        return StringUtils.join(str1, str2, str3, str4, str5);
    }

    /**
     * Implements {@code str:join(str1, str2, separator)} EL function - joins 2
     * strings with a separator.
     *
     * @param str1
     *            the first string.
     * @param str2
     *            the second string.
     * @param separator
     *            the separator string to use.
     *
     * @return the joined string.
     */
    public static String strJoin(final String str1, final String str2, final String separator)
    {
        return join(separator, str1, str2);
    }

    /**
     * Implements {@code str:join3(str1, str2, str3, separator)} EL function -
     * joins 3 strings with a separator.
     *
     * @param str1
     *            the first string.
     * @param str2
     *            the second string.
     * @param str3
     *            the third string.
     * @param separator
     *            the separator string to use.
     *
     * @return the joined string.
     */
    public static String strJoin3(final String str1, final String str2, final String str3,
            final String separator)
    {
        return join(separator, str1, str2, str3);
    }

    /**
     * Implements {@code str:join4(str1, str2, str3, str4, separator)} EL
     * function - joins 4 strings with a separator.
     *
     * @param str1
     *            the first string.
     * @param str2
     *            the second string.
     * @param str3
     *            the third string.
     * @param str4
     *            the fourth string.
     * @param separator
     *            the separator string to use.
     *
     * @return the joined string.
     */
    public static String strJoin4(final String str1, final String str2, final String str3,
            final String str4, final String separator)
    {
        return join(separator, str1, str2, str3, str4);
    }

    /**
     * Implements {@code str:join5(str1, str2, str3, str4, str5, separator)} EL
     * function - joins 5 strings with a separator.
     *
     * @param str1
     *            the first string.
     * @param str2
     *            the second string.
     * @param str3
     *            the third string.
     * @param str4
     *            the fourth string.
     * @param str5
     *            the fifth string.
     * @param separator
     *            the separator string to use.
     *
     * @return the joined string.
     */
    public static String strJoin5(final String str1, final String str2, final String str3,
            final String str4, final String str5, final String separator)
    {
        return join(separator, str1, str2, str3, str4, str5);
    }

    @Override
    public Method resolveFunction(final String prefix, final String localName)
    {
        if ("str".equals(prefix))
        {
            if ("concat".equals(localName))
            {
                return StrConcat.INSTANCE;
            }
            else if ("concat3".equals(localName))
            {
                return StrConcat3.INSTANCE;
            }
            else if ("concat4".equals(localName))
            {
                return StrConcat4.INSTANCE;
            }
            else if ("concat5".equals(localName))
            {
                return StrConcat5.INSTANCE;
            }
            else if ("join".equals(localName))
            {
                return StrJoin.INSTANCE;
            }
            else if ("join3".equals(localName))
            {
                return StrJoin3.INSTANCE;
            }
            else if ("join4".equals(localName))
            {
                return StrJoin4.INSTANCE;
            }
            else if ("join5".equals(localName))
            {
                return StrJoin5.INSTANCE;
            }
        }

        return null;
    }
}
